/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "server.h"
#include "util/memory.h"
#include "assert/assert.h"
#include "console/console.h"

server_t SERVER;

void serverInit() {
  memoryZero(&SERVER, sizeof(server_t));
  pthread_mutex_init(&SERVER.startMutex, NULL);
  pthread_cond_init(&SERVER.startCond, NULL);
}

errorret_t serverStart(uint16_t port) {
  assertFalse(SERVER.isRunning, "Server is already running?");
  consolePrint("Starting server on port %d...\n", port);

  pthread_mutex_lock(&SERVER.startMutex);
  SERVER.threadReady = false;

  // Initialize the server socket
  SERVER.serverSocket = socket(AF_INET, SOCK_STREAM, 0);
  if(SERVER.serverSocket < 0) {
    pthread_mutex_unlock(&SERVER.startMutex);
    return error("Failed to create socket");
  }
  SERVER.serverAddress.sin_family = AF_INET;
  SERVER.serverAddress.sin_addr.s_addr = INADDR_ANY;
  SERVER.serverAddress.sin_port = htons(port);

  // Bind the socket to the address and port
  int32_t res = bind(
    SERVER.serverSocket,
    (struct sockaddr *)&SERVER.serverAddress,
    sizeof(SERVER.serverAddress)
  );
  if(res < 0) {
    close(SERVER.serverSocket);
    pthread_mutex_unlock(&SERVER.startMutex);
    return error("Failed to bind socket");
  }

  // Set the socket to listen for incoming connections
  res = listen(SERVER.serverSocket, SERVER_MAX_CLIENTS);
  if(res < 0) {
    close(SERVER.serverSocket);
    pthread_mutex_unlock(&SERVER.startMutex);
    return error("Failed to listen on socket");
  }

  // Start the server thread
  SERVER.isRunning = true;
  if(pthread_create(&SERVER.thread, NULL, serverThread, NULL) != 0) {
    perror("Failed to create server thread");
    serverStop();
    pthread_mutex_unlock(&SERVER.startMutex);
    return error("Failed to create server thread");
  }

  // Wait for the server thread to be ready
  while(!SERVER.threadReady) {
    pthread_cond_wait(&SERVER.startCond, &SERVER.startMutex);
  }

  pthread_mutex_unlock(&SERVER.startMutex);
  consolePrint("Server started.");
  return ERROR_OK;
}

void * serverThread(void *arg) {
  struct timeval timeout;
  fd_set readfds;

  pthread_mutex_lock(&SERVER.startMutex);
  SERVER.threadReady = true;
  pthread_cond_signal(&SERVER.startCond);
  pthread_mutex_unlock(&SERVER.startMutex);

  while(SERVER.isRunning) {
    usleep(SERVER_TIMEOUT * 1000);

    FD_ZERO(&readfds);
    FD_SET(SERVER.serverSocket, &readfds);

    timeout.tv_usec = SERVER_TIMEOUT * 1000;

    // Wait for 
    int32_t activity = select(
      SERVER.serverSocket + 1,
      &readfds,
      NULL,
      NULL,
      &timeout
    );

    // Check for errors
    if(activity < 0) {
      consolePrint("Error in select");
      continue;
    }

    // Timeout
    if(activity == 0) continue;

    // Check if there is a new connection
    if(!FD_ISSET(SERVER.serverSocket, &readfds)) continue;

    // Accept new connection
    int32_t clientSocket = accept(SERVER.serverSocket, NULL, NULL);
    if(clientSocket < 0) {
      consolePrint("Failed to accept connection");
      continue;
    }


    // Initialize the client
    serverclient_t *client = NULL;
    uint8_t i = 0;
    do {
      if(SERVER.clients[i].state != SERVER_CLIENT_STATE_DISCONNECTED) {
        i++;
        continue;
      }
      client = &SERVER.clients[i];
      break;
    } while(i < SERVER_MAX_CLIENTS);

    // Can we receive this client?
    if(!client) {
      const char_t *errMsg = "ERR|Server full";
      send(clientSocket, errMsg, strlen(errMsg), 0);
      consolePrint("Server full, closing connection.");
      close(clientSocket);
      continue;
    }

    errorret_t ret = serverClientAccept(client, clientSocket);
    if(ret != ERROR_OK) {
      consolePrint("Failed to accept client connection", errorString());
      errorFlush();
      close(clientSocket);
    }

    usleep(1000);
  }

  printf("Server thread exiting.\n");
  assertFalse(
    SERVER.isRunning, "Server thread exiting while server is running?"
  );
  return NULL;
}

uint8_t serverGetClientCount() {
  uint8_t i = 0, clientCount = 0;
  do {
    if(SERVER.clients[i].state == SERVER_CLIENT_STATE_DISCONNECTED) continue;
    clientCount++;
  } while(i < SERVER_MAX_CLIENTS);
  return clientCount;
}

void serverStop() {
  consolePrint("Stopping server...");

  SERVER.isRunning = false;
  if(SERVER.serverSocket >= 0) {
    close(SERVER.serverSocket);
    SERVER.serverSocket = -1;
  }

  if(SERVER.thread) {
    pthread_join(SERVER.thread, NULL);
    SERVER.thread = 0;
  }

  pthread_mutex_destroy(&SERVER.startMutex);
  pthread_cond_destroy(&SERVER.startCond);

  consolePrint("Server stopped.");
}

void serverDispose() {
  serverStop();
}